# Flutter Mediator

[![Pub](https://img.shields.io/pub/v/flutter_mediator.svg)](https://pub.dev/packages/flutter_mediator)

Flutter mediator is a MVC state management package base on InheritedModel with aspects to make them simpler, easier, and intuitive to use.

<p align="center"><img  src="https://raw.githubusercontent.com/rob333/flutter_mediator/main/doc/images/main.gif"></p>

## Features

- **_Easy and simple to use_** - automatically make things work by only a few steps
- **_Efficiency and Performance_** - use InheritedModel as the underlying layer to make the app quick response
- **_Lightweight_** - small package size
- **_Flexible_** - provide both automatic and manual observation
- **_Intuitive_** - three main classes: **_`Publisher`, `Subscriber`, `Host`_**
  <br /> &emsp; **_`Publisher`_** - publish aspects
  <br /> &emsp; **_`Subscriber`_** - subscribe aspects
  <br /> &emsp; **_`Host`_** - place on top of the widget tree, dispatch aspects using InheritedModel

## Setting up

Add the following dependency to pubspec.yaml of your flutter project:

```yaml
dependencies:
  flutter_mediator: "^1.0.0"
```

Import flutter_mediator in files that will be used:

```dart
import 'package:flutter_mediator/mediator.dart';
```

For help getting started with Flutter, view the online [documentation](https://flutter.dev/docs).

## Getting Started Quick Steps: **_`Model`_**, **_`View`_**, **_`Controller`_**, and **_`Host`_**

### **_`Model`_**

1. Extend the model from **_`Publisher`_**
2. Declare rx variables with **_`.rx`_**, which will automatically rebuild the aspect-related-widget when updated.
3. Implement the update methods.

##### MyModel.dart

```dart
class MyModel extends Publisher {
  /// `.rx` make the var automatically rebuild the aspect-related-widget when updated
  var int1 =  0.rx;

  void updateInt1() {
    /// since int1 is a rx variable,
    /// it will automatically rebuild the aspect-realted-widget when updated
    int1 +=  1;
  }
```

### **_`Subscribe the View`_**

4. **_Subscribe the widget with designated one or more aspect._**
   <br /> 4-1. Choice an aspect to represent the widget.
   <br /> &emsp; &emsp; In this example, a string type aspect with value **_`'int1'`_** is designated.
   <br /> 4-2. Designate the model to be used.
   <br /> &emsp; &emsp; In this example, myModel of class **_`<MyModel>`_** is used.
   <br /> _Then any rx variables used inside the create method will automatically rebuild the aspect-related-widget when updated._ _(triggered by getter and setter)_

```dart
class Int1Listener extends StatelessWidget {
  @override
  Widget build(BuildContext context) {
    return 'int1'.subModel<MyModel>(
      create: (context, model) {
        /// Since model.int1 is a rx variable,
        /// it will automatically rebuild the aspect-related-widget when updated.
        /// In this example, the aspect is 'int1' of <MyModel>.
        return  Text('Int1 is ${model.int1}');
      },
    );
  }
}
```

### **_`Controller`_**

5. In the Widget tree, execute the update method.
   <br /> For example, get the model of class **_`<MyModel>`_** and execute it's update method within a RaisedButton.

```dart
RaisedButton(
  child: const Text('Update Int1'),
  onPressed: () => context.getModel<MyModel>().updateInt1(),
);
```

### **_`Host`_**

6. Inject **_host_** at the top level of the widget tree.

```dart
void  main() {
  runApp(
    MultiHost.create2(
      MyModel(updateMs:  1000), // model extends from Publisher
      ListModel(updateMs:  500),// model extends from Publisher
      child:  MyApp(),
    ),
  /// MultiHost.create1 to MultiHost.create9 are provided by the package.
  );
}
```

### **_`Works automatically!`_**

7. Then the state works automatically!

## Example

You can find the example in the [example](https://github.com/rob333/flutter_mediator/tree/main/example/lib) folder.
<br /> Try once, then you will see it's simple and easy to use.

## Detail

1.  [**single model**](#1-single-model) - host
2.  [**multiple models**](#2-multiple-models) - host
3.  [**automatically update the widget** with rx variable](#3-automatically-update-the-widget) - Publisher
4.  [**access the underlying value directly** of rx variable](#4-access-the-underlying-value-directly) - Publisher
5.  [**update the value by call style** of rx variable](#5-update-the-value-by-call-style) - Publisher
6.  [**manually publish an aspect** of the model](#6-manually-publish-an-aspect) - Publisher
7.  [**manually publish more than one aspect** of the model](#7-manually-publish-more-than-one-aspect) - Publisher
8.  [**manually publish all aspects** of the model](#8-manually-publish-all-aspects) - Publisher
9.  [**future publish** of the model](#9-future-publish) - Publisher
10. [**rebuild only once per frame** for the same aspect](#10-rebuild-only-once-per-frame) - Publisher
11. [**writing model helper**](#11-writing-model-helper) - Publisher
12. [**get model value**](#12-get-model-value) - Subscriber
13. [**subscribe a aspect** of the model](#13-subscribe-a-aspect) - Subscriber
14. [**subscribe more than one aspect** of the model](#14-subscribe-more-than-one-aspect) - Subscriber
15. [**subscribe all aspects** of the model](#15-subscribe-all-aspects) - Subscriber
16. [**subscribe with enum aspects** of the model](#16-subscribe-with-enum-aspects) - Subscriber
17. [**writing custom rx class**](#17-writing-custom-rx-class)

### 1. single model

```dart
void main() {
  runApp(
    Host(
      model: AppModel(), // model extends from Publisher
      child: MyApp(),
    ),
  );
}
```

&emsp; [back to detail](#detail)

### 2. multiple models

**_`MultiHost.create1`_** to **_`MultiHost.create9`_** are provided by the package.
<br /> You can add more `MultiHost.createXX` method, see [multi_host.dart](https://github.com/rob333/flutter_mediator/blob/main/lib/mediator/multi_host.dart) for example.

```dart
void main() {
  runApp(
    MultiHost.create2(
      MyModel(updateMs: 1000),  // model extends from Publisher
      ListModel(updateMs: 500), // model extends from Publisher
      child:  MyApp(),
    ),
  );
}
```

&emsp; [back to detail](#detail)

### 3. automatically update the widget

**_`.rx`_** make the var automatically rebuild the aspect-related-widget when updated

##### int:

```dart
class MyModel extends Publisher {
/// `.rx` make the var automatically update the aspect-related-widget
var int1 = 0.rx;

void  updateInt1() {
  int1 += 1; // automatically update the aspect-related-widget
}
```

##### List:

```dart
class ListModel extends Publisher {
  /// `.rx` make the var automatically update the aspect-related-widget
  final data =  <ListItem>[].rx;

  void updateListItem() {
    // get new item data...
    final newItem = ListItem(itemName, units, color);
    data.add(newItem); // automatically update the aspect-related-widget
  }
```

rx variable of type `int`, `double`, `num`, `string`, `bool`, `list`, `map`, `set` are provided by the package.
<br /> See also _[RxInt class](https://github.com/rob333/flutter_mediator/blob/main/lib/mediator/rx/rx_num.dart#L417)_,
[RxList class](https://github.com/rob333/flutter_mediator/blob/main/lib/mediator/rx/rx_iterable/rx_list.dart#L5),
[RxList.add](https://github.com/rob333/flutter_mediator/blob/main/lib/mediator/rx/rx_iterable/rx_list.dart#L35)

&emsp; [back to detail](#detail)

### 4. access the underlying value directly

Access the underlying value directly by `.value`.

##### MyModel.dart

```dart
var int1 = 0.rx;
void updateInt1() {
  // int1 += 1;
  /// is the same as
  int1.value += 1; // automatically update the aspect-related-widget
}
```

&emsp; [back to detail](#detail)

### 5. update the value by call style

##### MyModel.dart

```dart
var _foo = 1.rx;
set foo(int value) {
  _foo(value); // update rx variable by call() style
  /// is the same as
  // _foo = value;
  /// is the same as
  // _foo.value = value;
}
```

&emsp; [back to detail](#detail)

### 6. manually publish an aspect

##### MyModel.dart

```dart
int manuallyInt = 0;
void manuallyPublishDemo(int value) {
  manuallyInt = value;
  publish('manuallyInt'); // manually publish aspect of 'manuallyInt'
}
```

&emsp; [back to detail](#detail)

### 7. manually publish more than one aspect

Putting aspects in a list to publish more than one aspect.

##### MyModel.dart

```dart
int _foo = 0;
int _bar = 0;
void increaseBoth() {
  _foo += 1;
  _bar += 1;
  publish(['foo', 'bar']); // manually publish aspects in a list
}
```

&emsp; [back to detail](#detail)

### 8. manually publish all aspects

Publish null value to publish all aspects of the model.

##### MyModel.dart

```dart
void increaseAll() {
  //...
  publish(); // publish all aspects of the model
}
```

&emsp; [back to detail](#detail)

### 9. future publish

Use with a future method, for example,

##### MyModel.dart

```dart
int int1 = 0.rx;
Future<void> futureInt1() async {
  await Future.delayed(const Duration(seconds: 1));
  int1 += 1; // int1 is Rx, automatically update the aspect-related-widget
}
```

&emsp; [back to detail](#detail)

### 10. rebuild only once per frame

Rebuild only once per frame for the same aspect.
<br /> Following code only cause the aspect-related-widget to rebuild once.

##### MyModel.dart

```dart
int int1 = 0.rx;
void incermentInt1() async {
  int1 += 1;       // int1 is Rx, automatically update the aspect-related-widget
  publish('int1'); // manually publish 'int1'
  publish('int1'); // manually publish 'int1', again
  // only cause the related widgets to rebuild once per frame

}
```

&emsp; [back to detail](#detail)

### 11. writing model helper

You can write model helpers to simplified the typing. For example,

##### MyModel.dart

```dart
/// Helper function of MyModel
MyModel getMyModel(BuildContext context) {
  return Host.getInheritOfModel<MyModel>(context);
}

Subscriber<MyModel> subMyModel({Key key, Object aspects,
                @required CreaterOfSubscriber<MyModel> create}) {
  return aspects.subModel<MyModel>(key: key, create: create);
}

extension MyModelHelperT<T> on T {
  Subscriber<MyModel> subMyModel({Key key,
                @required CreaterOfSubscriber<MyModel> create}) {
    return Subscriber<MyModel>(key: key, aspects: this, create: create);
  }
}
```

##### ListModel.dart

```dart
/// Helper function of ListModel
ListModel getListModel(BuildContext context) {
  return Host.getInheritOfModel<ListModel>(context);
}

Subscriber<ListModel> subListModel({Key key, Object aspects,
                @required CreaterOfSubscriber<ListModel> create}) {
  return aspects.subModel<ListModel>(key: key, create: create);
}

extension ListModelHelperT<T> on T {
  Subscriber<ListModel> subListModel({Key key,
                @required CreaterOfSubscriber<ListModel> create}) {
    return Subscriber<ListModel>(key: key, aspects:  this, create: create);
  }
}
```

&emsp; [back to detail](#detail)

### 12. get model value

To get the model, for example, getting `MyModel`,

##### original form

```dart
final model = Host.getInheritOfModel<MyModel>(context);
```

##### by package helper of context extension

(_See [mediator_helper.dart](https://github.com/rob333/flutter_mediator/blob/main/lib/mediator/mediator_helper.dart#L25) for package helper._)

```dart
final model = context.getModel<MyModel>();
```

##### by model helper on your own

```dart
final model = getMyModel(context);
```

#### Get current triggered frame aspects of the model.

See also [AllListener@main.dart](https://github.com/rob333/flutter_mediator/blob/main/example/lib/main.dart#L176).

```dart
final model = context.getModel<MyModel>();
final aspects = model.frameAspect;
```

&emsp; [back to detail](#detail)

### 13. subscribe a aspect

For example, listen to a `String` type aspect **_`'int1'`_** of class **_`<MyModel>`_**.

##### original form

```dart
class Int1Listener extends StatelessWidget {
  @override
  Widget build(BuildContext context) {
    return Subscriber<MyModel>(
      aspects: 'int1',
      create: (context, model) {
        return Text('Int1 is ${model.int1}');
      },
    );
  }
}
```

##### with helper

```dart
class Int1Listener extends StatelessWidget {
  @override
  Widget build(BuildContext context) {
    // return 'int1'.subMyModel(     // with model helper on your own
    return 'int1'.subModel<MyModel>( // with package helper
      create: (context, model) {
        return Text('Int1 is ${model.int1}');
      },
    );
  }
}
```

&emsp; [back to detail](#detail)

### 14. subscribe more than one aspect

Putting aspects in a list to subscribe more than one aspect.

##### original form

```dart
class TwoListener extends StatelessWidget {
  @override
  Widget build(BuildContext context) {
    return Subscriber<MyModel>(
      aspects: ['int1', 'star'],
      create: (context, model) {
        return Text('Int1 is ${model.int1} and Star is ${model.star}');
      },
    );
  }
}
```

##### with helper

```dart
class TwoListener extends StatelessWidget {
  @override
  Widget build(BuildContext context) {
    // return ['int1', 'star'].subMyModel(     // with model helper on your own
    return ['int1', 'star'].subModel<MyModel>( // with package helper
      create: (context, model) {
        return Text('Int1 is ${model.int1} and Star is ${model.star}');
      },
    );
  }
}
```

&emsp; [back to detail](#detail)

### 15. subscribe all aspects

Provide no aspects parameter, or use null as aspect to listen to all aspects.
<br /> See also [AllListener@main.dart](https://github.com/rob333/flutter_mediator/blob/main/example/lib/main.dart#L176).

##### original form

```dart
class AllListener extends StatelessWidget {
  @override
  Widget build(BuildContext context) {
    return Subscriber<MyModel>(
      // aspects: , // no aspects parameter means listen to all aspects
      create: (context, model) {
        final aspects = model.frameAspect;
        final str = aspects == null ? '' : '$aspects received';
        return Text(str);
      },
    );
  }
}
```

##### with helper

```dart
class AllListener extends StatelessWidget {
  @override
  Widget build(BuildContext context) {
    // return null.subMyModel(     // with model helper on your own
    return null.subModel<MyModel>( // with package helper
      create: (context, model) {
        final aspects = model.frameAspect;
        final str = aspects == null ? '' : '$aspects received';
        return Text(str);
      },
    );
  }
}
```

&emsp; [back to detail](#detail)

### 16. subscribe with enum aspects

You can use `enum` as aspect type. For example, first, define the enum.

##### ListModel.dart

```dart
enum ListEnum {
  ListUpdate,
}
```

Then everything is the same as `String` type aspect, just to replace the `String` with `enum`.
<br /> See also [ListItemView@main.dart](https://github.com/rob333/flutter_mediator/blob/main/example/lib/main.dart#L108).

##### original form

```dart
class ListItemView extends StatelessWidget {
  @override
  Widget build(BuildContext context) {
    return Subscriber<ListModel>(
      aspects: ListEnum.ListUpdate,
      create: (context, model) {
      //...
    });
  }
}
```

##### with helper

```dart
class ListItemView extends StatelessWidget {
  @override
  Widget build(BuildContext context) {
    // return  ListEnum.ListUpdate.subListModel(create: (context, model) { // with model helper on your own
    return ListEnum.ListUpdate.subModel<ListModel>(create: (context, model) { // with package helper
      //...
    });
  }
}
```

&emsp; [back to detail](#detail)

### 17. writing custom rx class

If you need to write your own rx class, see [custom_rx_class.dart](https://github.com/rob333/flutter_mediator/blob/main/example/lib/custom_rx_class.dart) for example.

&emsp; [back to detail](#detail)

&nbsp;&nbsp;

## Changelog

Please see the [Changelog](https://github.com/rob333/flutter_mediator/blob/main/CHANGELOG.md) page.

## License

[MIT](https://github.com/rob333/flutter_mediator/blob/main/LICENSE)
